@@ -261,6 +261,9 @@ struct ucp_ep_config_key {
 
     /* Additional flags */
     unsigned                 flags;
+
+    /* Indicates peer release version */
+    unsigned                 dst_version;
 };
 
 
@@ -470,6 +473,12 @@ struct ucp_ep_config {
 
     /* Bitmap of preregistration for am_bw lanes */
     ucp_md_map_t                  am_bw_prereg_md_map;
+
+    /* Bitmap of lanes selected by the protocols */
+    ucp_lane_map_t                proto_lane_map;
+
+    /* Number of endpoints using this configuration */
+    unsigned                      ep_count;
 };
 
 
@@ -772,7 +781,8 @@ int ucp_ep_is_local_connected(ucp_ep_h ep);
 
 unsigned ucp_ep_local_disconnect_progress(void *arg);
 
-size_t ucp_ep_tag_offload_min_rndv_thresh(ucp_ep_config_t *config);
+size_t ucp_ep_tag_offload_min_rndv_thresh(ucp_context_h context,
+                                          const ucp_ep_config_key_t *key);
 
 void ucp_ep_config_rndv_zcopy_commit(ucp_lane_index_t lanes_count,
                                      ucp_ep_rndv_zcopy_config_t *rndv_zcopy);
@@ -881,4 +891,17 @@ ucs_status_t ucp_ep_query_sockaddr(ucp_ep_h ucp_ep, ucp_ep_attr_t *attr);
  */
 ucs_status_t ucp_ep_realloc_lanes(ucp_ep_h ep, unsigned new_num_lanes);
 
+
+/**
+ * @brief Set configuration index to the endpoint.
+ * 
+ * Changing of the configuration index deactivates UCP worker interfaces
+ * corresponding to the previous endpoint configuration and activates interfaces
+ * of the new configurtion.
+ *
+ * @param [in] ep         Endpoint object.
+ * @param [in] cfg_index  Endpoint configuration index.
+ */
+void ucp_ep_set_cfg_index(ucp_ep_h ep, ucp_worker_cfg_index_t cfg_index);
+
 #endif
